import uuid

from collections import defaultdict
from datetime import datetime
from decimal import Decimal

import barker.schemas.split as schemas

from fastapi import APIRouter, HTTPException, Security, status
from sqlalchemy import delete, func, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from ...core.security import get_current_active_user as get_user
from ...db.session import SessionFuture
from ...models.inventory import Inventory
from ...models.inventory_modifier import InventoryModifier
from ...models.kot import Kot
from ...models.overview import Overview
from ...models.settle_option import SettleOption
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...routers.voucher import (
    check_permissions,
    do_update_bill_numbers,
    do_update_settlements,
    do_update_table,
)
from ...schemas.receive_payment import ReceivePaymentItem as SettleSchema
from ...schemas.user_token import UserToken


router = APIRouter()


@router.post("/split-bill/{id_}")
def split(
    id_: uuid.UUID,
    data: schemas.Split,
    u: bool,  # Update table?
    user: UserToken = Security(get_user, scopes=["split-bill"]),
):
    try:
        with SessionFuture() as db:
            now = datetime.utcnow()
            update_table = u
            item: Voucher = db.execute(select(Voucher).where(Voucher.id == id_)).scalar_one()
            original_voucher_type = item.voucher_type
            item.voucher_type = VoucherType.VOID
            item.reason = "Bill Split"
            do_update_bill_numbers(item, db)
            do_update_settlements(item, [SettleSchema(id_=SettleOption.VOID(), amount=round(item.amount))], db)
            if update_table:
                db.execute(delete(Overview).where(Overview.voucher_id == item.id))
            check_permissions(None, item.voucher_type, user.permissions)

            one_inventories = [i for k in item.kots for i in k.inventories if i.id in data.inventories]
            two_inventories = [i for k in item.kots for i in k.inventories if i.id not in data.inventories]

            one = save(
                one_inventories,
                now,
                original_voucher_type,
                0,
                data.table_id,
                item.customer_id,
                user.id_,
                db,
            )
            if update_table:
                do_update_table(one, None, db)
            two = save(
                two_inventories,
                now,
                original_voucher_type,
                item.pax,
                item.food_table_id,
                item.customer_id,
                user.id_,
                db,
            )
            if update_table:
                do_update_table(two, None, db)
            db.commit()
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def save(
    inventories: list[Inventory],
    now: datetime,
    voucher_type: VoucherType,
    pax: int,
    table_id: uuid.UUID,
    customer_id: uuid.UUID | None,
    user_id: uuid.UUID,
    db: Session,
):
    if not are_product_quantities_positive(inventories):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Quantity of a product is negative",
        )
    if len(inventories) == 0:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="No inventories selected",
        )
    if happy_hour_items_more_than_regular(inventories):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="When product has happy hours\n"
            "Minimum same number of regular items also needed in the whole bill.",
        )
    kot_id = db.execute(select(func.coalesce(func.max(Voucher.kot_id), 0) + 1)).scalar_one()

    item: Voucher = Voucher(now, pax, kot_id, table_id, customer_id, voucher_type, user_id)
    db.add(item)
    for split_inventories in split_into_kots(inventories):
        if not happy_hour_items_balanced(split_inventories):
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Happy hour products are not balanced.",
            )
        code = db.execute(select(func.coalesce(func.max(Kot.code), 0) + 1)).scalar_one()
        kot = Kot(item.id, code, item.food_table_id, item.date, item.user_id)
        item.kots.append(kot)
        db.add(kot)
        db.flush()
        for old_inventory in split_inventories:
            inv = Inventory(
                kot.id,
                old_inventory.product_id,
                old_inventory.quantity,
                old_inventory.price,
                old_inventory.discount,
                old_inventory.is_happy_hour,
                old_inventory.tax_id,
                old_inventory.tax_rate,
                old_inventory.sort_order,
            )
            kot.inventories.append(inv)
            db.add(inv)
            for m in old_inventory.modifiers:
                mod = InventoryModifier(None, m.modifier_id, m.price)
                inv.modifiers.append(mod)
                db.add(mod)
    db.flush()
    do_update_bill_numbers(item, db)
    do_update_settlements(item, [], db)
    db.flush()
    return item


def split_into_kots(inventories: list[Inventory]) -> list:
    kots = defaultdict(list)
    for item in inventories:
        kots[item.kot_id].append(item)
    return [k for k in kots.values() if len(k) > 0]


def happy_hour_items_balanced(inventories: list[Inventory]) -> bool:
    happy = set((i.product_id, i.quantity) for i in inventories if i.is_happy_hour)
    products = set(i.product_id for i in inventories if i.is_happy_hour)
    other = set((i.product_id, i.quantity) for i in inventories if not i.is_happy_hour and i.product_id in products)
    return happy == other


def are_product_quantities_positive(inventories: list[Inventory]) -> bool:
    quantities: dict[tuple[uuid.UUID, bool], Decimal] = defaultdict(Decimal)
    for i in inventories:
        key = (i.product_id, i.is_happy_hour)
        quantities[key] += i.quantity
    for j in quantities.values():
        if j < 0:
            return False
    return True


def happy_hour_items_more_than_regular(invs: list[Inventory]) -> bool:
    inventories = {}
    for inventory in invs:
        if inventory.product_id not in inventories:
            inventories[inventory.product_id] = {"normal": Decimal(0), "happy": Decimal(0)}
        if inventory.is_happy_hour:
            inventories[inventory.product_id]["happy"] += inventory.quantity
        else:
            inventories[inventory.product_id]["normal"] += inventory.quantity
    for value in inventories.values():
        if value["happy"] > value["normal"]:
            return True
    return False
